package chat

import java.util.concurrent._
import zgs.httpd._

object ChatApp extends HLetExecutor {
  
  // config and commons -----------------------------------------------------------------------
  private val (truncateLength, truncatePeriod) = (300, 10000L)
  private val (schedulerSleepMillis, schedulerTimeout) = (1000L, (ChatServer.connectionTimeoutSeconds - 5) * 1000L)
  require(schedulerTimeout > 10000L) // to force some sense for long polling
  private def now = System.currentTimeMillis
  private def onError(e: Throwable): Unit = e.printStackTrace
  
  // the list of new submitted messages (in reverse order) with scheduled truncation ----------
  case class Msg(val id: Long, val stamp: Long, val nic: String, val text: String) {
    private def hms : String = new java.text.SimpleDateFormat("HH:mm:ss") format new java.util.Date(stamp)
    def toXml = <msg><id>{id}</id><stamp>{hms}</stamp><nic>{nic}</nic><text>{text}</text></msg>
  }
  
  private val nextId = new atomic.AtomicLong(1L)
  private var messages: List[Msg] = Nil
  private def syncMsg[T](code: => T): T = messages.synchronized { code }
  
  def acceptMsg(nic: String, text: String) = syncMsg { messages = Msg(nextId.getAndIncrement, now, nic, text) :: messages }
  def getMessages(afterId: Long): Seq[Msg] = syncMsg { messages.takeWhile(_.id > afterId).reverse }
  
  new java.util.Timer(/* daemon: */ true) scheduleAtFixedRate(
    new java.util.TimerTask {
      def run { 
        syncMsg { messages = messages.take(truncateLength) }
      }
    }, 
    truncatePeriod,
    truncatePeriod)

  // pollers - as a queue to be more fair -----------------------------------------------------
  private val pollers = new zgs.sync.SyncQu[Runnable]
  def submit(req: HReqData, r: Runnable): Unit = pollers.put(r) // implements HLetExecutor
  
  private val exe = new zgs.sync.SyncExe(
    poolSize = 10, // it's a matter of config for real app
    queueCap = Int.MaxValue,
    name = "PollersPool",
    onError = ChatApp.onError)
  
  // start scheduling -------------------------------------------------------------------------
  private def broadcast: Unit = pollers.drain.foreach(exe.submit(_))
  private val lastBroadcast = new atomic.AtomicLong(0L)
  private def needBroadcast = (pollers.size > 0) || (lastBroadcast.get + schedulerTimeout > now)
  
  zgs.sync.Sync.daemonNamed("ChatScheduler") {
    while (true) { 
      Thread.sleep(schedulerSleepMillis)
      if (needBroadcast) { 
        broadcast
        lastBroadcast.set(now)
      }
    }
  }
  
  //-------------------------------------------------------------------------------------------
  def stop: Unit = {
    def tryIt[T](code: => Unit): Unit = try { code } catch { case e => e.printStackTrace }
    tryIt(ChatApp.pollers.close)
    tryIt(ChatApp.exe.shutdown)
  }
}
